/**************************************************************************
*  Original Diku Mud copyright (C) 1990, 1991 by Sebastian Hammer,        *
*  Michael Seifert, Hans Henrik St{rfeldt, Tom Madsen, and Katja Nyboe.   *
*                                                                         *
*  Merc Diku Mud improvements copyright (C) 1992, 1993 by Michael         *
*  Chastain, Michael Quan, and Mitchell Tse.                              *
*                                                                         *
*  In order to use any part of this Merc Diku Mud, you must comply with   *
*  both the original Diku license in 'license.doc' as well the Merc       *
*  license in 'license.txt'.  In particular, you may not remove either of *
*  these copyright notices.                                               *
*                                                                         *
*  Much time and thought has gone into this software and you are          *
*  benefiting.  We hope that you share your changes too.  What goes       *
*  around, comes around.                                                  *
***************************************************************************
*       ROM 2.4 is copyright 1993-1998 Russ Taylor                        *
*       ROM has been brought to you by the ROM consortium                 *
*           Russ Taylor (rtaylor@hypercube.org)                           *
*           Gabrielle Taylor (gtaylor@hypercube.org)                      *
*           Brian Moore (zump@rom.org)                                    *
*       By using this code, you have agreed to follow the terms of the    *
*       ROM license, in the file Rom24/doc/rom.license                    *
***************************************************************************
*       1stMUD ROM Derivative (c) 2001-2003 by Ryan Jennings              *
*            http://1stmud.dlmud.com/  <r-jenn@shaw.ca>                   *
***************************************************************************/

#include "merc.h"
#include "interp.h"
#include "recycle.h"
#include "tables.h"

/*
 
 Note Board system, (c) 1995-96 Erwin S. Andreasen, erwin@andreasen.org
 =====================================================================
 
 Basically, the notes are split up into several boards. The boards do not
 exist physically, they can be read anywhere and in any position.
 
 Each of the note boards has its own file. Each of the boards can have its own
 "rights": who can read/write.
 
 Each character has an extra field added, namele the timestamp of the last note
 read by him/her on a certain board.
 
 The note entering system is changed too, making it more interactive. When
 entering a note, a character is put AFK and into a special CON_ state.
 Everything typed goes into the note.
 
 For the immortals it is possible to purge notes based on age. An Archive
 options is available which moves the notes older than X days into a special
 board. The file of this board should then be moved into some other directory
 during e.g. the startup script and perhaps renamed depending on date.
 
 Note that write_level MUST be >= read_level or else there will be strange
 output in certain functions.
 
 Board DEFAULT_BOARD must be at least readable by *everyone*.
 
*/

#define L_SUP (MAX_LEVEL - 1)	/* if not already defined */

BOARD_DATA boards[MAX_BOARD] = {

	{"Announce", "Announcements from Immortals", 0, L_SUP, "all", DEF_NORMAL,
	 60, NULL, NULL, FALSE}
	,
	{"General", "General discussion", 0, 2, "all", DEF_INCLUDE, 21, NULL, NULL,
	 FALSE}
	,
	{"Ideas", "Suggestion for improvement", 0, 2, "all", DEF_NORMAL, 60, NULL, NULL,
	 FALSE}
	,

	{"Bugs", "Typos, bugs, errors", 0, 1, "imm", DEF_NORMAL, 60, NULL, NULL,
	 FALSE}
	,
	{"Personal", "Personal messages", 0, 1, "all", DEF_EXCLUDE, 28, NULL, NULL,
	 FALSE}
	,

	{"Immortal", "Immortals only", LEVEL_IMMORTAL, LEVEL_IMMORTAL, "imm",
	 DEF_INCLUDE, 21, NULL, NULL, FALSE}

};

/* The prompt that the character is given after finishing a note with ~ or END */
const char *szFinishPrompt =
	"({WC{x)ontinue, ({WV{x)iew, ({WP{x)ost or ({WF{x)orget it?";

long last_note_stamp = 0;		/* To generate unique timestamps on notes */

NOTE_DATA *note_first;
NOTE_DATA *note_last;
NOTE_DATA *note_free;

#define BOARD_NOACCESS -1
#define BOARD_NOTFOUND -1

#undef EXTERN
#define EXTERN
PROTOTYPE(static bool next_board, (CHAR_DATA *));
#undef EXTERN
#define EXTERN EXTERN

/* recycle a note */
void free_note(NOTE_DATA * note)
{
	if (note->sender)
		free_string(note->sender);
	if (note->to_list)
		free_string(note->to_list);
	if (note->subject)
		free_string(note->subject);
	if (note->date)				/* was note->datestamp for some reason */
		free_string(note->date);
	if (note->text)
		free_string(note->text);

	PUT_FREE(note, next, note_free);
}

/* allocate memory for a new note or recycle */
NOTE_DATA *new_note()
{
	NOTE_DATA *note;

	GET_FREE(note, NOTE_DATA, next, note_free);

	/* Zero all the field - Envy does not gurantee zeroed memory */
	note->next = NULL;
	note->sender = NULL;
	note->expire = 0;
	note->to_list = NULL;
	note->subject = NULL;
	note->date = NULL;
	note->date_stamp = 0;
	note->text = NULL;

	return note;
}

/* append this note to the given file */
static void append_note(FILE * fp, NOTE_DATA * note)
{
	fprintf(fp, "Sender  %s~\n", note->sender);
	fprintf(fp, "Date    %s~\n", note->date);
	fprintf(fp, "Stamp   %ld\n", note->date_stamp);
	fprintf(fp, "Expire  %ld\n", note->expire);
	fprintf(fp, "To      %s~\n", note->to_list);
	fprintf(fp, "Subject %s~\n", note->subject);
	fprintf(fp, "Text\n%s~\n\n", note->text);
}

/* Save a note in a given board */
void finish_note(BOARD_DATA * board, NOTE_DATA * note)
{
	WRITE_DATA *fp;
	char filename[200];

	/* The following is done in order to generate unique date_stamps */

	if (last_note_stamp >= current_time)
		note->date_stamp = ++last_note_stamp;
	else
	{
		note->date_stamp = current_time;
		last_note_stamp = current_time;
	}

	LINK(note, board->note_first, board->note_last, next, prev);

	/* append note to note file */

	sprintf(filename, "%s%s", NOTE_DIR, board->short_name);

	fp = open_append(filename);
	if (!fp)
	{
		bug("Could not open one of the note files in append mode");
		board->changed = TRUE;	/* set it to TRUE hope it will be OK later? */
		return;
	}

	append_note(fp->stream, note);
	close_append(fp);
}

/* Find the number of a board */
int board_number(const BOARD_DATA * board)
{
	int i;

	for (i = 0; i < MAX_BOARD; i++)
		if (board == &boards[i])
			return i;

	return -1;
}

/* Find a board number based on  a string */
int board_lookup(const char *name)
{
	int i;

	for (i = 0; i < MAX_BOARD; i++)
		if (!str_cmp(boards[i].short_name, name))
			return i;

	return -1;
}

/* Remove list from the list. Do not free note */
static void unlink_note(BOARD_DATA * board, NOTE_DATA * note)
{
	UNLINK(note, board->note_first, board->note_last, next, prev);
}

/* Find the nth note on a board. Return NULL if ch has no access to that note */
static NOTE_DATA *find_note(CHAR_DATA * ch, BOARD_DATA * board, int num)
{
	int count = 0;
	NOTE_DATA *p;

	for (p = board->note_first; p; p = p->next)
		if (++count == num)
			break;

	if ((count == num) && is_note_to(ch, p))
		return p;
	else
		return NULL;

}

/* save a single board */
static void save_board(BOARD_DATA * board)
{
	WRITE_DATA *fp;
	char filename[200];
	char buf[200];
	NOTE_DATA *note;

	sprintf(filename, "%s%s", NOTE_DIR, board->short_name);

	fp = open_write(filename);

	if (!fp)
	{
		sprintf(buf, "Error writing to: %s", filename);
		bug(buf);
	}
	else
	{
		for (note = board->note_first; note; note = note->next)
			append_note(fp->stream, note);

		close_write(fp);
	}
}

/* Show one not to a character */
static void show_note_to_char(CHAR_DATA * ch, NOTE_DATA * note, int num)
{

	/* Ugly colors ? */
	chprintlnf(ch,
			   "[{W%4d{x] {Y%s{x: {g%s{x\n\r" "{YDate{x:  %s\n\r"
			   "{YTo{x:    %s\n\r"
			   "{g==========================================================================={x\n\r"
			   "%s", num, note->sender, note->subject, note->date,
			   note->to_list, note->text);
}

/* Save changed boards */
void save_notes()
{
	int i;

	for (i = 0; i < MAX_BOARD; i++)
		if (boards[i].changed)	/* only save changed boards */
			save_board(&boards[i]);
}

/* Load a single board */
static void load_board(BOARD_DATA * board)
{
	READ_DATA *fp;
	WRITE_DATA *fp_archive;
	char filename[200];

	sprintf(filename, "%s%s", NOTE_DIR, board->short_name);

	fp = open_read(filename);

	/* Silently return */
	if (!fp)
		return;

	/* Start note fetching. copy of db.c:load_notes() */

	for (;;)
	{
		NOTE_DATA *pnote;
		char letter;

		do
		{
			if ((letter = read_letter(fp)) == NUL)
			{
				close_read(fp);
				return;
			}
		}
		while (isspace(letter));
		fp->pos--;

		alloc_mem(pnote, NOTE_DATA, 1);

		if (str_cmp(read_word(fp), "sender"))
			break;
		pnote->sender = read_string(fp);

		if (str_cmp(read_word(fp), "date"))
			break;
		pnote->date = read_string(fp);

		if (str_cmp(read_word(fp), "stamp"))
			break;
		pnote->date_stamp = read_number(fp);

		if (str_cmp(read_word(fp), "expire"))
			break;
		pnote->expire = read_number(fp);

		if (str_cmp(read_word(fp), "to"))
			break;
		pnote->to_list = read_string(fp);

		if (str_cmp(read_word(fp), "subject"))
			break;
		pnote->subject = read_string(fp);

		if (str_cmp(read_word(fp), "text"))
			break;
		pnote->text = read_string(fp);

		/* Should this note be archived right now ? */

		if (pnote->expire < current_time)
		{
			char archive_name[200];

			sprintf(archive_name, "%s%s.old", NOTE_DIR, board->short_name);
			fp_archive = open_append(archive_name);
			if (!fp_archive)
				bug("Could not open archive boards for writing");
			else
			{
				append_note(fp_archive->stream, pnote);
				close_append(fp_archive);	/* it might be more efficient to close this later */
			}

			free_note(pnote);
			board->changed = TRUE;
			continue;

		}

		LINK(pnote, board->note_first, board->note_last, next, prev);
	}

	bug("Load_notes: bad key word.");
	return;						/* just return */
}

/* Initialize structures. Load all boards. */
void load_boards()
{
	int i;

	for (i = 0; i < MAX_BOARD; i++)
		load_board(&boards[i]);
}

/* Returns TRUE if the specified note is address to ch */
bool is_note_to(CHAR_DATA * ch, NOTE_DATA * note)
{
	if (!str_cmp(ch->name, note->sender))
		return TRUE;

	if (is_ignoring(ch, note->to_list, IGNORE_NOTES))
		return FALSE;

	if (is_full_name("all", note->to_list))
		return TRUE;

	if (IS_IMMORTAL(ch) &&
		(is_full_name("imm", note->to_list) ||
		 is_full_name("imms", note->to_list) ||
		 is_full_name("immortal", note->to_list) ||
		 is_full_name("god", note->to_list) ||
		 is_full_name("gods", note->to_list) ||
		 is_full_name("immortals", note->to_list)))
		return TRUE;

	if ((get_trust(ch) == MAX_LEVEL) &&
		(is_full_name("imp", note->to_list) ||
		 is_full_name("imps", note->to_list) ||
		 is_full_name("implementor", note->to_list) ||
		 is_full_name("implementors", note->to_list)))
		return TRUE;

	if (is_full_name(ch->name, note->to_list))
		return TRUE;

	/* Allow a note to e.g. 40 to send to characters level 40 and above */
	if (is_number(note->to_list) && get_trust(ch) >= atoi(note->to_list))
		return TRUE;

	if (is_clan(ch) && (is_full_name(ch->clan->name, note->to_list)
						|| is_full_name("clan", note->to_list)))
		return TRUE;

	return FALSE;
}

/* Return the number of unread notes 'ch' has in 'board' */
/* Returns BOARD_NOACCESS if ch has no access to board */
int unread_notes(CHAR_DATA * ch, BOARD_DATA * board)
{
	NOTE_DATA *note;
	time_t last_read;
	int count = 0;

	if (board->read_level > get_trust(ch))
		return BOARD_NOACCESS;

	last_read = ch->pcdata->last_note[board_number(board)];

	for (note = board->note_first; note; note = note->next)
		if (is_note_to(ch, note) &&
			((long) last_read < (long) note->date_stamp))
			count++;

	return count;
}

/*
 * COMMANDS
 */

/* Start writing a note */
static void do_nwrite(CHAR_DATA * ch, const char *argument)
{
	char buf[200];

	if (IS_NPC(ch))				/* NPC cannot post notes */
		return;

	if (get_trust(ch) < ch->pcdata->board->write_level)
	{
		chprintln(ch, "You cannot post notes on this board.");
		return;
	}

	/* continue previous note, if any text was written */
	if (ch->pcdata->in_progress && (!ch->pcdata->in_progress->text))
	{
		chprintln
			(ch,
			 "Note in progress cancelled because you did not manage to write any text \n\r"
			 "before losing link.\n\r");
		free_note(ch->pcdata->in_progress);
		ch->pcdata->in_progress = NULL;
	}

	if (!ch->pcdata->in_progress)
	{
		ch->pcdata->in_progress = new_note();
		replace_string(ch->pcdata->in_progress->sender, ch->name);

		replace_string(ch->pcdata->in_progress->date,
					   str_time(current_time, -1, NULL));
	}

	act("{G$n starts writing a note.{x", ch, NULL, NULL, TO_ROOM);

	/* Begin writing the note ! */
	chprintlnf(ch,
			   "You are now %s a new note on the {W%s{x board.\n\r"
			   "If you are using tintin, type #verbose to turn off alias expansion!\n\r",
			   ch->pcdata->in_progress->text ? "continuing" : "posting",
			   ch->pcdata->board->short_name);

	chprintlnf(ch, "{YFrom{x:    %s\n\r", ch->name);

	if (!ch->pcdata->in_progress->text)	/* Are we continuing an old note or not? */
	{
		switch (ch->pcdata->board->force_type)
		{
		case DEF_NORMAL:
			sprintf(buf,
					"If you press Return, default recipient \"{W%s{x\" will be chosen.\n\r",
					ch->pcdata->board->names);
			break;
		case DEF_INCLUDE:
			sprintf(buf,
					"The recipient list MUST include \"{W%s{x\". If not, it will be added automatically.\n\r",
					ch->pcdata->board->names);
			break;

		case DEF_EXCLUDE:
			sprintf(buf,
					"The recipient of this note must NOT include: \"{W%s{x\".",
					ch->pcdata->board->names);

			break;
		}

		chprintln(ch, buf);
		chprintln(ch, "{YTo{x:      ");

		ch->desc->connected = CON_NOTE_TO;
		/* nanny takes over from here */

	}
	else						/* we are continuing, print out all the fields and the note so far */
	{
		chprintlnf(ch,
				   "{YTo{x:      %s\n\r" "{YExpires{x: %s\n\r"
				   "{YSubject{x: %s", ch->pcdata->in_progress->to_list,
				   str_time(ch->pcdata->in_progress->expire, GET_TZONE(ch),
							NULL), ch->pcdata->in_progress->subject);
		chprintln(ch, "{GYour note so far:{x");
		chprint(ch, ch->pcdata->in_progress->text);

		chprintlnf
			(ch,
			 "\n\rEnter text. Type {W%cq{x or {W@{x on an empty line to end note, or {W%ch{x for help.\n\r"
			 "======================================================================\n\r",
			 STR_EDIT_KEY(ch), STR_EDIT_KEY(ch));

		ch->desc->connected = CON_NOTE_TEXT;

	}

}

/* Read next note in current group. If no more notes, go to next board */
static void do_nread(CHAR_DATA * ch, const char *argument)
{
	NOTE_DATA *p;
	int count = 0, number;
	time_t *last_note = &ch->pcdata->last_note[board_number(ch->pcdata->board)];

	if (!str_cmp(argument, "again"))
	{							/* read last note again */

	}
	else if (is_number(argument))
	{
		number = atoi(argument);

		for (p = ch->pcdata->board->note_first; p; p = p->next)
			if (++count == number)
				break;

		if (!p || !is_note_to(ch, p))
			chprintln(ch, "No such note.");
		else
		{
			show_note_to_char(ch, p, count);
			*last_note = UMAX(*last_note, p->date_stamp);
		}
	}
	else						/* just next one */
	{

		count = 1;
		for (p = ch->pcdata->board->note_first; p; p = p->next, count++)
			if ((p->date_stamp > *last_note) && is_note_to(ch, p))
			{
				show_note_to_char(ch, p, count);
				/* Advance if new note is newer than the currently newest for that char */
				*last_note = UMAX(*last_note, p->date_stamp);
				return;
			}

		chprintln(ch, "No new notes in this board.");

		if (next_board(ch))
			chprintlnf(ch, "Changed to next board, %s.",
					   ch->pcdata->board->short_name);
		else
			chprintln(ch, "There are no more boards.");
	}
}

/* Remove a note */
static void do_nremove(CHAR_DATA * ch, const char *argument)
{
	NOTE_DATA *p;

	if (!is_number(argument))
	{
		chprintln(ch, "Remove which note?");
		return;
	}

	p = find_note(ch, ch->pcdata->board, atoi(argument));
	if (!p)
	{
		chprintln(ch, "No such note.");
		return;
	}

	if (str_cmp(ch->name, p->sender) && (get_trust(ch) < MAX_LEVEL))
	{
		chprintln(ch, "You are not authorized to remove this note.");
		return;
	}

	unlink_note(ch->pcdata->board, p);
	free_note(p);
	chprintln(ch, "Note removed!");

	save_board(ch->pcdata->board);	/* save the board */
}

/* List all notes or if argument given, list N of the last notes */
/* Shows REAL note numbers! */
static void do_nlist(CHAR_DATA * ch, const char *argument)
{
	int count = 0, show = 0, num = 0, has_shown = 0;
	time_t last_note;
	NOTE_DATA *p;

	if (is_number(argument))	/* first, count the number of notes */
	{
		show = atoi(argument);

		for (p = ch->pcdata->board->note_first; p; p = p->next)
			if (is_note_to(ch, p))
				count++;
	}

	chprintln(ch, "{WNotes on this board:{x\n\r"
			  "{rNum> Author        Subject{x");

	last_note = ch->pcdata->last_note[board_number(ch->pcdata->board)];

	for (p = ch->pcdata->board->note_first; p; p = p->next)
	{
		num++;
		if (is_note_to(ch, p))
		{
			has_shown++;		/* note that we want to see X VISIBLE note, not just last X */
			if (!show || ((count - show) < has_shown))
			{
				chprintf(ch, "{W%3d{x>{B%c{Y%-13s{y %s{x",
						 num,
						 last_note < p->date_stamp ? '*' : ' ',
						 p->sender, p->subject);
			}
		}

	}
}

/* catch up with some notes */
static void do_ncatchup(CHAR_DATA * ch, const char *argument)
{
	NOTE_DATA *p;

	/* Find last note */
	for (p = ch->pcdata->board->note_first; p && p->next; p = p->next);

	if (!p)
		chprintln(ch, "Alas, there are no notes in that board.");
	else
	{
		ch->pcdata->last_note[board_number(ch->pcdata->board)] = p->date_stamp;
		chprintln(ch, "All mesages skipped.");
	}
}

CH_CMD(do_ncheck)
{
	int i, count = 0, unread = 0;

	for (i = 0; i < MAX_BOARD; i++)
	{
		unread = unread_notes(ch, &boards[i]);
		if (unread != BOARD_NOACCESS)
			count += unread;
	}

	if (count < 1)
		chprintln(ch, "You have no new notes on the board.");
	else
		chprintlnf(ch,
				   "You have {Y%d{x unread note%s on the board. Type 'board'.",
				   count, (count != 1) ? "s" : "");
}

/* Dispatch function for backwards compatibility */
CH_CMD(do_note)
{
	char arg[MAX_INPUT_LENGTH];

	if (IS_NPC(ch))
		return;

	argument = one_argument(argument, arg);

	if ((!arg[0]) || (!str_cmp(arg, "read")))	/* 'note' or 'note read X' */
		do_nread(ch, argument);

	else if (!str_cmp(arg, "list"))
		do_nlist(ch, argument);

	else if (!str_cmp(arg, "write"))
		do_nwrite(ch, argument);

	else if (!str_cmp(arg, "remove"))
		do_nremove(ch, argument);

	else if (!str_cmp(arg, "purge"))
		chprintln(ch, "Obsolete.");

	else if (!str_cmp(arg, "archive"))
		chprintln(ch, "Obsolete.");

	else if (!str_cmp(arg, "catchup"))
		do_ncatchup(ch, argument);
	else if (!str_cmp(arg, "check"))
		do_ncheck(ch, argument);
	else
		do_oldhelp(ch, "note");
}

/* Show all accessible boards with their numbers of unread messages OR
   change board. New board name can be given as a number or as a name (e.g.
    board personal or board 4 */
CH_CMD(do_board)
{
	int i, count, number;

	if (IS_NPC(ch))
		return;

	if (!argument[0])			/* show boards */
	{
		int unread;

		count = 1;
		chprintln(ch, "{RNum         Name Unread Description{x\n\r"
				  "{r=== ============ ====== ==========={x");
		for (i = 0; i < MAX_BOARD; i++)
		{
			unread = unread_notes(ch, &boards[i]);	/* how many unread notes? */
			if (unread != BOARD_NOACCESS)
			{
				chprintlnf(ch,
						   "{W%2d{x> {G%12s{x [%s%4d{x] {y%s{x",
						   count, boards[i].short_name,
						   unread ? "{r" : "{g", unread, boards[i].long_name);
				count++;
			}					/* if has access */

		}						/* for each board */

		chprintlnf(ch, "\n\rYou current board is {W%s{x.",
				   ch->pcdata->board->short_name);

		/* Inform of rights */
		if (ch->pcdata->board->read_level > get_trust(ch))
			chprintln(ch, "You cannot read nor write notes on this board.");
		else if (ch->pcdata->board->write_level > get_trust(ch))
			chprintln(ch, "You can only read notes from this board.");
		else
			chprintln(ch, "You can both read and write on this board.");

		return;
	}							/* if empty argument */

	if (ch->pcdata->in_progress)
	{
		chprintln(ch, "Please finish your interrupted note first.");
		return;
	}

	/* Change board based on its number */
	if (is_number(argument))
	{
		count = 0;
		number = atoi(argument);
		for (i = 0; i < MAX_BOARD; i++)
			if (unread_notes(ch, &boards[i]) != BOARD_NOACCESS)
				if (++count == number)
					break;

		if (count == number)	/* found the board.. change to it */
		{
			ch->pcdata->board = &boards[i];
			chprintlnf(ch, "Current board changed to {W%s{x. %s.",
					   boards[i].short_name,
					   (get_trust(ch) <
						boards[i].write_level) ?
					   "You can only read here" :
					   "You can both read and write here");
		}
		else					/* so such board */
			chprintln(ch, "No such board.");

		return;
	}

	/* Non-number given, find board with that name */

	for (i = 0; i < MAX_BOARD; i++)
		if (!str_cmp(boards[i].short_name, argument))
			break;

	if (i == MAX_BOARD)
	{
		chprintln(ch, "No such board.");
		return;
	}

	/* Does ch have access to this board? */
	if (unread_notes(ch, &boards[i]) == BOARD_NOACCESS)
	{
		chprintln(ch, "No such board.");
		return;
	}

	ch->pcdata->board = &boards[i];
	chprintlnf(ch, "Current board changed to {W%s{x. %s.",
			   boards[i].short_name,
			   (get_trust(ch) <
				boards[i].write_level) ? "You can only read here" :
			   "You can both read and write here");
}

/* Send a note to someone on the personal board */
void personal_message(const char *sender, const char *to,
					  const char *subject, const int expire_days,
					  const char *text)
{
	make_note("Personal", sender, to, subject, expire_days, text);
}

void make_note(const char *board_name, const char *sender, const char *to,
			   const char *subject, const int expire_days, const char *text)
{
	int board_index = board_lookup(board_name);
	BOARD_DATA *board;
	NOTE_DATA *note;

	if (board_index == BOARD_NOTFOUND)
	{
		bug("make_note: board not found");
		return;
	}

	if (strlen(text) > MAX_NOTE_TEXT)
	{
		bugf("make_note: text too long (%d bytes)", strlen(text));
		return;
	}

	board = &boards[board_index];

	note = new_note();			/* allocate new note */

	note->sender = str_dup(sender);
	note->to_list = str_dup(to);
	note->subject = str_dup(subject);
	note->expire = current_time + expire_days * 60 * 60 * 24;
	note->text = str_dup(text);

	note->date = str_dup(str_time(current_time, -1, NULL));

	finish_note(board, note);

}

/* tries to change to the next accessible board */
static bool next_board(CHAR_DATA * ch)
{
	int i = board_number(ch->pcdata->board) + 1;

	while ((i < MAX_BOARD) && (unread_notes(ch, &boards[i]) == BOARD_NOACCESS))
		i++;

	if (i == MAX_BOARD)
		return FALSE;
	else
	{
		ch->pcdata->board = &boards[i];
		return TRUE;
	}
}

void handle_con_note_to(DESCRIPTOR_DATA * d, const char *argument)
{
	char buf[MAX_INPUT_LENGTH];
	CHAR_DATA *ch = d->character;

	if (!ch->pcdata->in_progress)
	{
		d->connected = CON_PLAYING;
		bugf("nanny: In CON_NOTE_TO, but no note in progress");
		return;
	}

	strcpy(buf, argument);
	smash_tilde(buf);			/* change ~ to - as we save this field as a string later */

	switch (ch->pcdata->board->force_type)
	{
	case DEF_NORMAL:			/* default field */
		if (!buf[0])			/* empty string? */
		{
			replace_string(ch->pcdata->in_progress->to_list,
						   ch->pcdata->board->names);
			d_printlnf(d, "Assumed default recipient: {W%s{x",
					   ch->pcdata->board->names);
		}
		else
			replace_string(ch->pcdata->in_progress->to_list, buf);

		break;

	case DEF_INCLUDE:			/* forced default */
		if (!is_full_name(ch->pcdata->board->names, buf))
		{
			strcat(buf, " ");
			strcat(buf, ch->pcdata->board->names);
			replace_string(ch->pcdata->in_progress->to_list, buf);

			d_printlnf(d,
					   "\n\rYou did not specify %s as recipient, so it was automatically added.\n\r"
					   "{YNew To{x :  %s", ch->pcdata->board->names,
					   ch->pcdata->in_progress->to_list);
		}
		else
			replace_string(ch->pcdata->in_progress->to_list, buf);
		break;

	case DEF_EXCLUDE:			/* forced exclude */
		if (!buf[0])
		{
			d_print(d, "You must specify a recipient.\n\r" "{YTo{x:      ", 0);
			return;
		}

		if (is_full_name(ch->pcdata->board->names, buf))
		{
			d_printf(d,
					 "You are not allowed to send notes to %s on this board. Try again.\n\r"
					 "{YTo{x:      ", ch->pcdata->board->names);
			return;				/* return from nanny, not changing to the next state! */
		}
		else
			replace_string(ch->pcdata->in_progress->to_list, buf);
		break;

	}

	d_print(d, "{Y\n\rSubject{x: ", 0);
	d->connected = CON_NOTE_SUBJECT;
}

void handle_con_note_subject(DESCRIPTOR_DATA * d, const char *argument)
{
	char buf[MAX_INPUT_LENGTH];
	CHAR_DATA *ch = d->character;

	if (!ch->pcdata->in_progress)
	{
		d->connected = CON_PLAYING;
		bugf("nanny: In CON_NOTE_SUBJECT, but no note in progress");
		return;
	}

	strcpy(buf, argument);
	smash_tilde(buf);			/* change ~ to - as we save this field as a string later */

	/* Do not allow empty subjects */

	if (!buf[0])
	{
		d_println(d, "Please find a meaningful subject!", 0);
		d_print(d, "{YSubject{x: ", 0);
	}
	else if (strlen(buf) > 60)
	{
		d_println(d,
				  "No, no. This is just the Subject. You're note writing the note yet. Twit.",
				  0);
	}
	else
		/* advance to next stage */
	{
		replace_string(ch->pcdata->in_progress->subject, buf);
		if (IS_IMMORTAL(ch))	/* immortals get to choose number of expire days */
		{
			d_printf(d,
					 "\n\rHow many days do you want this note to expire in?\n\r"
					 "Press Enter for default value for this board, {W%d{x days.\n\r"
					 "{YExpire{x:  ", ch->pcdata->board->purge_days);
			d->connected = CON_NOTE_EXPIRE;
		}
		else
		{
			ch->pcdata->in_progress->expire =
				current_time + ch->pcdata->board->purge_days * 24L * 3600L;
			d_printlnf(d, "This note will expire %s",
					   str_time(ch->pcdata->in_progress->expire, GET_TZONE(ch),
								NULL));
			d_printlnf(d,
					   "\n\rEnter text. Type {W%cq{x or {W@{x on an empty line to end note, or {W%ch{x for help.\n\r"
					   "======================================================================\n\r",
					   STR_EDIT_KEY(ch), STR_EDIT_KEY(ch));
			d->connected = CON_NOTE_TEXT;
		}
	}
}

void handle_con_note_expire(DESCRIPTOR_DATA * d, const char *argument)
{
	CHAR_DATA *ch = d->character;
	char buf[MAX_STRING_LENGTH];
	time_t expire;
	int days;

	if (!ch->pcdata->in_progress)
	{
		d->connected = CON_PLAYING;
		bugf("nanny: In CON_NOTE_EXPIRE, but no note in progress");
		return;
	}

	/* Numeric argument. no tilde smashing */
	strcpy(buf, argument);
	if (!buf[0])				/* assume default expire */
		days = ch->pcdata->board->purge_days;
	else /* use this expire */ if (!is_number(buf))
	{
		d_println(d, "Write the number of days!", 0);
		d_print(d, "{YExpire{x:  ", 0);
		return;
	}
	else
	{
		days = atoi(buf);
		if (days <= 0)
		{
			d_println(d,
					  "This is a positive MUD. Use positive numbers only! :)",
					  0);
			d_print(d, "{YExpire{x:  ", 0);
			return;
		}
	}

	expire = current_time + (days * 24L * 3600L);	/* 24 hours, 3600 seconds */

	ch->pcdata->in_progress->expire = expire;

	d_printlnf(d,
			   "\n\rEnter text. Type {W%cq{x or {W@{x on an empty line to end note, or {W%ch{x for help.\n\r"
			   "======================================================================\n\r",
			   STR_EDIT_KEY(ch), STR_EDIT_KEY(ch));
	d->connected = CON_NOTE_TEXT;
}

void handle_con_note_text(DESCRIPTOR_DATA * d, const char *argument)
{
	strshow_t action;
	CHAR_DATA *ch = d->character;
	char buf[MAX_STRING_LENGTH];

	if (!ch->pcdata->in_progress)
	{
		d->connected = CON_PLAYING;
		bugf("nanny: In CON_NOTE_TEXT, but no note in progress");
		return;
	}

	action = parse_string_command(&ch->pcdata->in_progress->text, argument, ch);

	switch (action)
	{
	case STRING_END:
		d_printf(d, "\n\r%s", szFinishPrompt);
		d->connected = CON_NOTE_FINISH;
		return;
	case STRING_FOUND:
		return;
	default:
	case STRING_NONE:
		smash_tilde(buf);		/* smash it now */

		if (ch->pcdata->in_progress->text)
		{
			strcpy(buf, ch->pcdata->in_progress->text);
			replace_string(ch->pcdata->in_progress->text, "");
		}
		else
			strcpy(buf, "");

		if ((strlen(argument) + strlen(buf)) > MAX_NOTE_TEXT)
		{
			d_println(d, "Note too long, bailing out!", 0);
			free_note(ch->pcdata->in_progress);
			ch->pcdata->in_progress = NULL;
			d->connected = CON_PLAYING;
			return;
		}

		strcat(buf, argument);
		strcat(buf, "\n\r");
		replace_string(ch->pcdata->in_progress->text, buf);
		return;
	}
}

void handle_con_note_finish(DESCRIPTOR_DATA * d, const char *argument)
{

	CHAR_DATA *ch = d->character;

	if (!ch->pcdata->in_progress)
	{
		d->connected = CON_PLAYING;
		bugf("nanny: In CON_NOTE_FINISH, but no note in progress");
		return;
	}

	switch (tolower(argument[0]))
	{
	case 'c':					/* keep writing */
		d_println(d, "Continuing note...", 0);
		d->connected = CON_NOTE_TEXT;
		break;

	case 'v':					/* view note so far */
		if (ch->pcdata->in_progress->text)
		{
			d_println(d, "{gText of your note so far:{x", 0);
			d_print(d, ch->pcdata->in_progress->text, 0);
		}
		else
			d_println(d, "You haven't written a thing!\n\r", 0);
		d_println(d, szFinishPrompt, 0);
		break;

	case 'p':					/* post note */
		finish_note(ch->pcdata->board, ch->pcdata->in_progress);
		d_println(d, "Note posted.", 0);
		d->connected = CON_PLAYING;
		announce(ch, INFO_NOTE,
				 "You have a new note on the %s board from $n",
				 ch->pcdata->board->short_name);
		/* remove AFK status */
		ch->pcdata->in_progress = NULL;
		act("{G$n finishes $s note.{x", ch, NULL, NULL, TO_ROOM);
		break;

	case 'f':
		d_println(d, "Note cancelled!", 0);
		free_note(ch->pcdata->in_progress);
		ch->pcdata->in_progress = NULL;
		d->connected = CON_PLAYING;
		/* remove afk status */
		break;

	default:					/* invalid response */
		d_println(d, "Huh? Valid answers are:\n\r", 0);
		d_println(d, szFinishPrompt, 0);
	}
}
